#include <string.h>
#include <fstream>
#include <list>
#include "csprng.h"
using namespace std;

csprng::csprng() {
	initLists();
	seed1 = new char[SEED_LEN];
	seed2 = new char[SEED_LEN];
	generateSeed(seed1, SEED_LEN);
	generateSeed(seed2, SEED_LEN);
	initSeed();
}

csprng::csprng(char* seed, unsigned int count) {
	initLists();
	this->seed(seed, count);
}

void csprng::initLists() {
	last1 = new list<char>();
	last2 = new list<char>();
}

csprng::~csprng() {
	delete last1;
	delete last2;
	delete seed1;
	delete seed2;
}

void csprng::generateSeed(char* seed, unsigned int count) {
	fstream* fs = new fstream();
	fs->open("/dev/urandom", fstream::in | fstream::binary);
	fs->read(seed, count);
	fs->close();
	delete fs;
}

void csprng::seed(char* seed, unsigned int count) {
	seed1 = new char[SEED_LEN];
	seed2 = new char[SEED_LEN];
	if (count >= SEED_LEN << 1) {
		memcpy(seed1, seed, SEED_LEN);
		memcpy(seed2, seed + SEED_LEN, SEED_LEN);
	} else {
		const int count1 = count > 1 ? count >> 1 : count;
		int i = 0;
		do {
			int c = SEED_LEN - i;
			memcpy(seed1 + i, seed, c < count1 ? c : count1);
			i += count1;
		} while (i < SEED_LEN);
		const int count2 = count > 1 ? count % 2 + count1 : count;
		const int offset = count > 1 ? count1 : 0;
		i = 0;
		do {
			int c = SEED_LEN - i;
			memcpy(seed2 + i, seed + offset, c < count2 ? c : count2);
			i += count2;
		} while (i < SEED_LEN);
	}
	initSeed();
}

inline void csprng::initSeed() {
	seedIndex = 0;
	last1->assign(seed1, seed1 + SEED_LEN);
	last2->assign(seed2, seed2 + SEED_LEN);
}

inline void rotate(list<char>* list) {
	list->push_back(list->front());
	list->pop_front();
}

void csprng::getRandom(void* result, unsigned int pSize) {
	for (int i = 0; i < pSize; ++i) {
		if ((last1->front() ^ last2->front()) % 2 == 0)
			rotate(last1);
		else
			rotate(last2);
		((char*)result)[i] = last1->front() ^ last2->front() ^ seed1[seedIndex] ^ seed2[seedIndex];
		if (++seedIndex >= SEED_LEN)
			seedIndex = 0;
	}
}
